from functools import total_ordering
import json
from logging import getLogger
from math import floor
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, TYPE_CHECKING
from urllib.parse import quote_plus

from IPython.display import display, HTML, Markdown
from markdown_strings import esc_format as escape, bold, link
import regex
from tqdm import tqdm

from ..util import normalize_ocr, cache, tokenize, strip
from ..config import CONFIG as _CONFIG

if TYPE_CHECKING:  # avoid circular dependency
    from .window import WindowPosition


CONFIG = _CONFIG['document.Document']
LOGGER = getLogger(__name__)


LeftContext = str
Match = str
RightContext = str
Snippet = Tuple[LeftContext, Match, RightContext]


@total_ordering
class Document:
    ROOT_PATH = Path(CONFIG['root_path'])
    RELEVANT_PAGES_PATH = Path(CONFIG['relevant_pages_path'])
    RELEVANT_PAGES = {
        int(book_id): set(pages)
        for book_id, pages
        in json.load(RELEVANT_PAGES_PATH.open('rt')).items()
    }
    SNIPPET_SIZE = CONFIG.getint('snippet_size')
    DOCUMENT_URL = CONFIG['document_url']
    IMAGE_URL = CONFIG['image_url']
    MAXIMUM_ERRORS = CONFIG.getint('maximum_errors')

    def __init__(self, filename: str):
        self.filename = filename
        self.basename = str(Path(filename).with_suffix(''))
        self.book_id, self.page_id = map(int, self.basename.split('/'))
        self.document_url = self.DOCUMENT_URL.format(book_id=self.book_id, page_id=self.page_id)
        self.image_url = self.IMAGE_URL.format(book_id=self.book_id, page_id=self.page_id)
        self.text_filename = self.ROOT_PATH / Path(self.filename).with_suffix('.txt')

    @staticmethod
    def get_basename(book_id: int, page_id: int) -> str:
        basename = f'{book_id}/{page_id}'
        return basename

    def get_snippet_url(self, position: 'WindowPosition') -> str:
        start, end = position
        snippet = self[start:end]
        tokens = tokenize(strip(snippet))
        value = quote_plus(' '.join(tokens))
        snippet_url = '{}&search={}'.format(self.document_url, value)
        return snippet_url

    def get_snippet(self, position: 'WindowPosition') -> Snippet:
        start, end = position
        assert start >= 0 and start <= len(self)
        assert end >= 0 and end <= len(self) and end >= start
        snippet_size = end - start
        if snippet_size < self.SNIPPET_SIZE:
            effective_start = max(0, start - floor(0.5 * (self.SNIPPET_SIZE - snippet_size)))
            snippet_size = end - effective_start
            effective_end = min(len(self), end + self.SNIPPET_SIZE - snippet_size)
        else:
            effective_start, effective_end = start, end

        left_context = self[effective_start:start]
        match = self[start:end]
        right_context = self[end:effective_end]

        if start != 0:
            left_context = '...{}'.format(left_context)
        if end != len(str(self)):
            right_context = '{}...'.format(right_context)

        snippet = (left_context, match, right_context)
        return snippet

    def find_snippet(self, snippet: Snippet) -> 'WindowPosition':
        position = self._find_exact_snippet(snippet)
        if position is None:
            position = self._find_tokenized_snippet(snippet)
        if position is None:
            raise ValueError(f'Snippet {snippet} not found in document {repr(self)}')
        return position

    def _find_exact_snippet(self, snippet: Snippet) -> Optional['WindowPosition']:
        left_context, match, right_context = snippet
        left_context = left_context.lstrip('...')
        right_context = right_context.rstrip('...')
        context = '{}{}{}'.format(left_context, match, right_context)

        if context not in str(self):
            return None

        start = str(self).index(context) + len(left_context)
        end = start + len(match)

        position = (start, end)
        return position

    def _find_tokenized_snippet(self, snippet: Snippet) -> Optional['WindowPosition']:
        left_context, match, right_context = snippet

        def create_full_match_regex(text: str) -> str:
            tokens = tokenize(text)
            token_regexes = map(regex.escape, tokens)
            full_match_regex = r'\W*'.join(token_regexes)
            return full_match_regex

        left_context_regex = create_full_match_regex(left_context)
        match_regex = create_full_match_regex(match)
        right_context_regex = create_full_match_regex(right_context)

        def _create_page_boundary_regex(text: str) -> Optional[str]:
            tokens = tokenize(text)
            if len(tokens) == 0:
                return None
            page_boundary_regex = f'(\\w+\\W*){{0,{len(tokens)}}}'
            return page_boundary_regex

        def create_page_beginning_regex(text: str) -> Optional[str]:
            page_boundary_regex = _create_page_boundary_regex(text)
            if page_boundary_regex is None:
                return None
            page_beginning_regex = f'^\\W*{page_boundary_regex}'
            return page_beginning_regex

        def create_page_end_regex(text: str) -> Optional[str]:
            page_boundary_regex = _create_page_boundary_regex(text)
            if page_boundary_regex is None:
                return None
            page_beginning_regex = f'{page_boundary_regex}\\W*$'
            return page_beginning_regex

        left_context_page_beginning_regex = create_page_beginning_regex(left_context)
        right_context_page_end_regex = create_page_end_regex(right_context)

        if left_context_page_beginning_regex is not None:
            left_context_regex = f'({left_context_page_beginning_regex}|{left_context_regex})'
        if right_context_page_end_regex is not None:
            right_context_regex = f'({right_context_page_end_regex}|{right_context_regex})'

        context_regex = f'{left_context_regex}\\W*(?<result>{match_regex})\\W*{right_context_regex}'
        match = regex.search(context_regex, str(self))

        if match is None and self.MAXIMUM_ERRORS > 0:
            context_regex = f'(?b)({context_regex}){{e<={self.MAXIMUM_ERRORS}}}'
            match = regex.search(context_regex, str(self))

        if match is None:
            return None

        start = match.start('result')
        end = match.end('result')

        position = (start, end)
        return position

    def get_markdown_snippet(self, position: 'WindowPosition') -> Markdown:
        left_context, match, right_context = self.get_snippet(position)
        snippet = '{}{}{}'.format(escape(left_context), bold(escape(match)), escape(right_context))
        return Markdown(snippet)

    @property
    @cache
    def text(self) -> str:
        with self.text_filename.open('rt') as f:
            text = f.read()
        return normalize_ocr(text)

    @classmethod
    @cache
    def get_filenames(cls, only_relevant: bool = False) -> Tuple[str]:
        num_total, num_relevant = 0, 0
        filenames = []
        input_filenames = cls.ROOT_PATH.glob('*/*.txt')
        if only_relevant:
            input_filenames = tqdm(list(input_filenames), desc='Filtering documents')
        for filename in input_filenames:
            filename = str(filename.relative_to(cls.ROOT_PATH))
            document = Document(filename)
            num_total += 1
            if only_relevant and document.book_id not in cls.RELEVANT_PAGES:
                continue
            if only_relevant and document.page_id not in cls.RELEVANT_PAGES[document.book_id]:
                continue
            num_relevant += 1
            filenames.append(filename)
        message = 'Found {} relevant documents out of all {} ({:.2f}%).'
        message = message.format(num_relevant, num_total, 100.0 * num_relevant / num_total)
        if only_relevant:
            LOGGER.info(message)
        return tuple(sorted(filenames))

    def __len__(self) -> int:
        return len(str(self))

    def __getitem__(self, key: Any) -> str:
        return str(self)[key]

    def __hash__(self):
        return hash(self.basename)

    def __eq__(self, other) -> bool:
        if isinstance(other, Document):
            return self.basename == other.basename
        return NotImplemented

    def __lt__(self, other) -> bool:
        if isinstance(other, Document):
            return self.basename < other.basename
        return NotImplemented

    def __str__(self) -> str:
        return self.text

    def __repr__(self) -> str:
        return self.basename

    def _ipython_display_(self) -> None:
        document_name = 'Document {}'.format(escape(self.basename))
        document_link = link(document_name, self.document_url)
        document_link = bold('{}:'.format(document_link))
        display(
            Markdown('{} {}'.format(document_link, escape(str(self)))),
            HTML('<img src="{}" width="500" />'.format(self.image_url)),
        )


def document_factory(filename: str) -> Tuple[str, Document]:
    document = Document(filename)
    return (document.basename, document)


def load_documents() -> Dict[str, Document]:
    filenames = Document.get_filenames()
    documents = {
        basename: document
        for basename, document
        in map(document_factory, tqdm(filenames, desc='Loading documents'))
    }
    return documents
